# standard libraries
import re
import typing as tp

# third party libraries
import requests
from bs4 import BeautifulSoup, Tag
from pydantic import BaseModel

from spelling_bee_cheat.data_model.puzzle import Puzzle

PUZZLE_URL = "https://www.nytimes.com/puzzles/spelling-bee"
GAME_DATA_RE = re.compile(r"window.gameData\s*=\s*(?P<data>{.*})")
WORDS_POINTS_PANGRAMS_RE = re.compile(r"WORDS: (?P<words>\d+), POINTS: (?P<points>\d+), PANGRAMS: (?P<pangrams>\d+)")


def search_tag(tag: Tag, subtag_type: str, regex: re.Pattern, description: str) -> tp.Tuple[Tag, re.Match]:
    """
    Search a tag for a subtag containing text matching a regex

    Args:
        tag: root tag to search
        subtag_type: type of tag to look for
        regex: pattern describing the target subtag
        description: description of what is being search for, to be added to
            an exception message if the search fails

    Returns:
        Tag: The first tag containing matching text
        re.Match: the match object from the re library

    Raises:
        RuntimeError: if a matching subtag is not found
    """
    for subtag in tag.find_all(subtag_type):
        mo = regex.search(subtag.text)
        if mo is not None:
            return subtag, mo
    raise RuntimeError(f"Could not find the {subtag_type} tag containing the {description}")


class TodayAndYesterdayPuzzles(BaseModel):
    """
    Today's puzzle data and yesterday's, just as they are bundled in the page
    data on the NYT spelling bee website.
    """

    today: Puzzle
    yesterday: Puzzle

    @classmethod
    def fetch(cls) -> "TodayAndYesterdayPuzzles":
        """
        Factory method for creating a new TodayAndYesterdayPuzzles instance.
        Fetches the game page from the NYT website, then extracts the puzzle
        data JSON.

        Returns:
            [TodayAndYesterdayPuzzles]: Game data scraped representing puzzles
            from today and yesterday
        """
        # Fetch the webpage that contains today's puzzle.
        r = requests.get(PUZZLE_URL)
        soup = BeautifulSoup(r.text, "html.parser")

        # Extract the JSON game data.
        _, mo = search_tag(soup, subtag_type="script", regex=GAME_DATA_RE, description="game data")
        game_data_json = mo.group("data")

        # Construct and return a pydantic representation.
        return cls.parse_raw(game_data_json)


def fetch_hint_text(puzzle: Puzzle) -> str:
    """
    Get hint text from the same day as the provided puzzle. The text giving the
    number of words/points/pangrams is returned. I intend to use this function
    to determine what the cannonically available hints are for spelling bee. I
    want to have the hints be generated from the puzzle data instead of scraping
    them and storing them.

    Args:
        puzzle: puzzle data corresponding to the desired hint data. The
        printDate attribute is used to get the hints from the same date.

    Returns:
        [str]: Text in the same HTML block as the words/points/pangrams counts
    """
    # Fetch the hints webpage.
    date_part = puzzle.printDate.replace("-", "/")
    url = f"https://www.nytimes.com/{date_part}/crosswords/spelling-bee-forum.html"
    r = requests.get(url)
    soup = BeautifulSoup(r.text, "html.parser")

    # Extract the number of words, points, and pangrams.
    p, _ = search_tag(
        soup,
        subtag_type="p",
        regex=WORDS_POINTS_PANGRAMS_RE,
        description="number of words, points, and pangrams",
    )

    # Return the raw text in the tag.
    return p.text



# standard libraries
from typing import List

from spelling_bee_cheat.dictionary import load_default_dictionary


def dictionary_search(
    letters: str,
    center_letter: str,
    dictionary: List[str] = None,
) -> List[str]:
    """
    Filter a list of words to those that satisfy the rules of the spelling bee game:
     - contains only letters from the 7 given
     - contains the letter at the center of the puzzle at least once
     - has a minimum length (usually 4)

    Args:
        letters: a string containing the seven allowed letters
        center_letter: a string containing the letter in the center of the puzzle
        dictionary: a list of words to consider. Defaults to nltk.corpus.words.words()

    Returns:
        List[str]: a list of words from the dictionary list that obey the rules of the game
    """
    if dictionary is None:
        dictionary = load_default_dictionary()

    letter_set = set(letters.lower())
    center_letter = center_letter.lower()

    words_found = [word for word in dictionary if center_letter in word and set(word) <= letter_set]

    return words_found



# standard libraries
import typing as tp
from pathlib import Path

from spelling_bee_cheat.data_model.puzzle import Puzzle
from spelling_bee_cheat.folders import PUZZLE_ARCHIVE_DIR
from spelling_bee_cheat.scrape.scrape import TodayAndYesterdayPuzzles

Pathlike = str | Path


class Archive:
    """
    Class to manage an archive of old spelling bee puzzles.
    """

    def __init__(self, path: Pathlike = PUZZLE_ARCHIVE_DIR):
        """
        Initializer for Archive class

        Args:
            path: directory in which to archive puzzles
        """
        self.path = Path(path)
        self.path.mkdir(parents=True, exist_ok=True)

    def path_from_iso8601(self, date_iso8601: str) -> Path:
        """
        Return the storage location for a puzzle from the puzzle date.

        Args:
            date_iso8601: a date string like "2022-02-26"
        """
        return self.path / f"{date_iso8601}.json"

    def puzzle_from_iso8601(self, date_iso8601: str) -> Puzzle | None:
        """
        Look for an archived puzzle from the requested date. Return None if not
        found. Expects date to be formatted according to ISO 8601.

        Args:
            date_iso8601: a date string like "2022-02-26"
        """
        in_path = self.path_from_iso8601(date_iso8601)
        if not in_path.exists():
            return None
        with in_path.open("r") as f:
            return Puzzle.parse_raw(f.read())

    def archive(self, puzzle: Puzzle) -> bool:
        """
        Save a Puzzle in the directory as JSON. Choose a file name from the
        publication date of the puzzle.

        Args:
            puzzle: puzzle to save

        Returns:
            bool: True if this puzzle was already in the archive, False otherwise
        """
        out_path = self.path_from_iso8601(puzzle.printDate)
        if not out_path.exists():
            # Avoid overwriting old data because of the missing expiration
            # value in the puzzle from the previous day.
            with out_path.open("w") as f:
                f.write(puzzle.json())
            return False
        return True

    def today_and_yesterday_puzzles(
        self, *, return_existed: bool = False
    ) -> TodayAndYesterdayPuzzles | tp.Tuple[TodayAndYesterdayPuzzles, bool, bool]:
        """
        Fetch and archive the puzzles from today and yesterday.

        Args:
            return_existed: if True, return the results of the calls to archive()

        Returns:
            TodayAndYesterdayPuzzles: object containing both puzzles
            bool: if today's puzzle already existed in the archive
            bool: if yesterday's puzzle already existed in the archive
        """
        tayp = TodayAndYesterdayPuzzles.fetch()
        today_existed = self.archive(tayp.today)
        yesterday_existed = self.archive(tayp.yesterday)
        if return_existed:
            return tayp, today_existed, yesterday_existed
        return tayp

    def today_puzzle(self) -> Puzzle:
        """
        Fetch and archive the puzzles from today and yesterday, returning only
        today's puzzle.

        Returns:
            Puzzle: today's puzzle
        """
        tayp = self.today_and_yesterday_puzzles()
        return tayp.today

    def yesterday_puzzle(self) -> Puzzle:
        """
        Fetch and archive the puzzles from today and yesterday, returning only
        yesterday's puzzle.

        Returns:
            Puzzle: yesterday's puzzle
        """
        tayp = self.today_and_yesterday_puzzles()
        return tayp.yesterday



#!/usr/bin/python
# standard libraries
import pickle

# third party libraries
import click

import spelling_bee_cheat


@click.command()
def convert_archive():
    """
    Convert the pickle files in the archive directory to json
    """
    archive = spelling_bee_cheat.archive.Archive()
    pickle_files = sorted(archive.path.glob("*.pkl"))
    for pf in pickle_files:
        with pf.open("rb") as f:
            puzzle = pickle.load(f)
        tayp = spelling_bee_cheat.scrape.TodayAndYesterdayPuzzles.parse_raw(puzzle.game_data_json)
        archive.archive(tayp.today)
        archive.archive(tayp.yesterday)


def main():
    """
    Main method. Call into click command
    """
    convert_archive()


if __name__ == "__main__":
    main()



# standard libraries
import typing as tp
from itertools import product
from string import ascii_lowercase

# third party libraries
import numpy as np
import pandas as pd
from more_itertools import chunked
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import make_pipeline
from tqdm import tqdm

from spelling_bee_cheat.dictionary import load_default_dictionary
from spelling_bee_cheat.nyt_scraping import PuzzleData

ascii_lowercase_list = list(ascii_lowercase)


class MarkovSearch:
    """ """

    @classmethod
    def _make_word(cls, g, l):  # noqa: E741
        return "".join(g.choice(ascii_lowercase_list, size=l))

    @classmethod
    def _make_negative_word(cls, g, l, positive_words):  # noqa: E741
        while True:
            word = cls._make_word(g, l)
            if word not in positive_words:
                return word

    @classmethod
    def _make_negative_words(cls, positive_words, g):
        positive_word_set = set(positive_words)
        return tuple(
            cls._make_negative_word(g, len(pword), positive_word_set)
            for pword in tqdm(positive_words, desc="making negative words")
        )

    @classmethod
    def _make_dataset(cls, random_state):
        g = np.random.default_rng(seed=random_state)

        positive_examples = load_default_dictionary()
        negative_examples = cls._make_negative_words(positive_examples, g)

        return pd.concat(
            (
                pd.DataFrame({"word": positive_examples, "exists": True}),
                pd.DataFrame({"word": negative_examples, "exists": False}),
            )
        )

    def __init__(self, random_state=1234):
        """ """
        self.pipeline = make_pipeline(
            TfidfVectorizer(analyzer="char_wb", ngram_range=(1, 3)),
            LogisticRegression(C=0.1),
        )

        self.dataset = self._make_dataset(random_state)

        self.pipeline.fit(self.dataset["word"], self.dataset["exists"])

    def markov_search(self, pdat: PuzzleData, batch_size: int = 100_000) -> tp.List[str]:
        """ """
        valid_letters = pdat.valid_letters
        center_letter = pdat.center_letter
        max_length = pd.Series(pdat.answers).str.len().max()

        words = []
        for length in range(4, max_length + 1):
            for letters in product(*[valid_letters] * length):
                if center_letter in letters:
                    words.append("".join(letters))

        dfuncs = []
        batch_iterable = tqdm(
            list(chunked(words, batch_size)),
            desc=f"Scoring words (batches of {batch_size})",
        )
        for words_chunk in batch_iterable:
            dfuncs.extend(self.pipeline.decision_function(words_chunk))

        self.words_df = pd.DataFrame(
            {
                "word": words,
                "dfunc": dfuncs,
            }
        )
        return self.words_df.sort_values("dfunc", ascending=False)["word"].to_list()



# standard libraries
import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple, Union

# third party libraries
import pandas as pd
import requests
from bs4 import BeautifulSoup, Tag

Pathlike = Union[str, Path]

PUZZLE_URL = "https://www.nytimes.com/puzzles/spelling-bee"
GAME_DATA_RE = re.compile(r"window.gameData\s*=\s*(?P<data>{.*})")
WORDS_POINTS_PANGRAMS_RE = re.compile(r"WORDS: (?P<words>\d+), POINTS: (?P<points>\d+), PANGRAMS: (?P<pangrams>\d+)")
SIGMA_RE = re.compile(r"Σ")
TWO_LETTER_GRID_RE = re.compile(r"([A-Z]{2}-\d+\s*)+")


def search_tag(tag: Tag, subtag_type: str, regex: re.Pattern, description: str) -> Tuple[Tag, re.Match]:
    """
    Search a tag for a subtag containing text matching a regex

    Args:
        tag: root tag to search
        subtag_type: type of tag to look for
        regex: pattern describing the target subtag
        description: description of what is being search for, to be added to
            an exception message if the search fails

    Returns:
        Tag: The first tag containing matching text
        re.Match: the match object from the re library

    Raises:
        RuntimeError: if a matching subtag is not found
    """
    for subtag in tag.find_all(subtag_type):
        mo = regex.search(subtag.text)
        if mo is not None:
            return subtag, mo
    raise RuntimeError(f"Could not find the {subtag_type} tag containing the {description}")


def text_with_br(tag: Tag) -> str:
    """
    Get text from a tag with br tags considered to be newlines. The replacement
    is destructive.

    Args:
        tag: a Tag from BeautifulSoup

    Returns:
        str: text from the tag, with br tags replaced by newlines
    """
    for br in tag.find_all("br"):
        br.replace_with("\n")
    return tag.text


def int_or_id(x: str) -> Union[int, str]:
    """
    Convert a string to int if possible

    Args:
        x: input string

    Returns:
        Union[int, str]: int represented by x, or x itself if parsing fails
    """
    try:
        return int(x)
    except ValueError:
        return x


@dataclass
class HintData:
    """
    Class representing the hints for a spelling bee puzzle
    """

    words: int
    points: int
    pangrams: int

    word_length_grid: pd.DataFrame
    two_letter_grid: pd.DataFrame

    @staticmethod
    def build_word_length_grid(text: str) -> pd.DataFrame:
        """
        Create a dataframe representation of the word length grid from the text
        representation found on the hints page. Output includes the sum column
        and sum row. Dashes are converted to zeros.

        Args:
            text: text from the webpage

        Returns:
            pd.DataFrame: dataframe with letters as index and lengths as columns
        """
        text = text.replace("-", "0")
        lines = text.split("\n")
        header, body = lines[0], lines[1:]

        columns = [int_or_id(x) for x in header.split()]
        index = []
        data = []
        for row in body:
            index_value, data_values = row.split(":")
            index.append(index_value)
            data.append([int(x) for x in data_values.split()])

        return pd.DataFrame(data, columns=columns, index=index)

    @staticmethod
    def build_two_letter_grid(text: str) -> pd.DataFrame:
        """
        Create a dataframe representation of the two letter list from the text
        representation found on the hints page. The representation from the page
        is described as a "two letter list," but this is rearranged into a
        table.

        Args:
            text: text from the webpage

        Returns:
            pd.DataFrame: dataframe with first letters as index and second
                letters as columns
        """
        two_letter_grid = pd.DataFrame()
        for two_letter_hint in text.split():
            letters, number = two_letter_hint.split("-")
            letter1, letter2 = letters
            number = int(number)
            two_letter_grid.loc[letter1, letter2] = number
        two_letter_grid.fillna(0, inplace=True, downcast="infer")
        sorted_columns = sorted(two_letter_grid.columns)
        two_letter_grid = two_letter_grid[sorted_columns]
        return two_letter_grid

    @classmethod
    def fetch_from_date(cls, date: str) -> "HintData":
        """
        Factory for new HintData instance from a date string. Builds a URL,
        fetches the hint page at that URL, and extracts data from it.

        Args:
            date: iso format date string
        """
        # Fetch the hints webpage.
        date_part = date.replace("-", "/")
        url = f"https://www.nytimes.com/{date_part}/crosswords/spelling-bee-forum.html"
        r = requests.get(url)
        soup = BeautifulSoup(r.text, "html.parser")

        # Extract the number of words, points, and pangrams.
        _, mo = search_tag(
            soup,
            subtag_type="p",
            regex=WORDS_POINTS_PANGRAMS_RE,
            description="number of words, points, and pangrams",
        )
        words, points, pangrams = map(int, mo.group("words", "points", "pangrams"))

        # Extract the word length grid.
        p, _ = search_tag(soup, subtag_type="p", regex=SIGMA_RE, description="word length grid")
        p_text = SIGMA_RE.sub(repl="sum", string=text_with_br(p))
        word_length_grid = cls.build_word_length_grid(p_text)

        # Extract the two letter list.
        p, _ = search_tag(
            soup,
            subtag_type="p",
            regex=TWO_LETTER_GRID_RE,
            description="two letter list",
        )
        p_text = text_with_br(p)
        two_letter_grid = cls.build_two_letter_grid(p_text)

        return cls(
            words=words,
            points=points,
            pangrams=pangrams,
            word_length_grid=word_length_grid,
            two_letter_grid=two_letter_grid,
        )


@dataclass
class PuzzleData:
    """
    Class representing a spelling bee puzzle
    """

    game_data_json: str
    game_data: dict

    date: str
    center_letter: str
    outer_letters: List[str]
    valid_letters: List[str]
    pangrams: List[str]
    answers: List[str]
    id_: int

    hint_data: HintData

    @classmethod
    def fetch_today(cls) -> "PuzzleData":
        """
        Factory method for creating a new PuzzleData instance. Fetches the game
        page from the NYT website, then extracts the puzzle data JSON.
        """
        # Fetch the webpage that contains today's puzzle.
        r = requests.get(PUZZLE_URL)
        soup = BeautifulSoup(r.text, "html.parser")

        # Extract the game data.
        _, mo = search_tag(soup, subtag_type="script", regex=GAME_DATA_RE, description="game data")
        game_data_json = mo.group("data")
        game_data = json.loads(game_data_json)

        # Extract fields from game data.
        today = game_data["today"]
        date = today["printDate"]
        center_letter = today["centerLetter"]
        outer_letters = today["outerLetters"]
        valid_letters = today["validLetters"]
        pangrams = today["pangrams"]
        answers = today["answers"]
        id_ = today["id"]

        # Use the date from the game data to load the hints.
        hint_data = HintData.fetch_from_date(date)

        return cls(
            game_data_json=game_data_json,
            game_data=game_data,
            date=date,
            center_letter=center_letter,
            outer_letters=outer_letters,
            valid_letters=valid_letters,
            pangrams=pangrams,
            answers=answers,
            id_=id_,
            hint_data=hint_data,
        )



#!/usr/bin/python
# third party libraries
import click

import spelling_bee_cheat


def report_action_taken(puzzle: spelling_bee_cheat.data_model.Puzzle, existed: bool) -> None:
    """
    Print summary of action taken with given Puzzle

    Args:
        puzzle: puzzle just archived
        existed: return value of the archive operation
    """
    if existed:
        print(f"Puzzle with date {puzzle.printDate} already existed, keeping old version")
    else:
        print("Archived", puzzle.printDate)


@click.command()
@click.option(
    "-d",
    "--archive-dir",
    type=click.Path(exists=True, file_okay=False, writable=True, readable=True),
    default=spelling_bee_cheat.folders.PUZZLE_ARCHIVE_DIR,
)
def scrape(archive_dir: str):
    """
    Download today's and yesterday's spelling bee puzzles from the NYT website
    and save them in JSON files. The file names will contain the publication
    date. They will look like YYYY-MM-DD.json. You can control the archive
    directory with a flag, but the default is the puzzle_archive/ directory in
    the code repo.

    Args:
        archive_dir: folder in which to save the results
    """
    archive = spelling_bee_cheat.archive.Archive(archive_dir)
    tayp, today_existed, yesterday_existed = archive.today_and_yesterday_puzzles(return_existed=True)
    report_action_taken(tayp.today, today_existed)
    report_action_taken(tayp.yesterday, yesterday_existed)


if __name__ == "__main__":
    scrape()



# standard libraries
import re
import typing as tp
from functools import cache

# third party libraries
import nltk
import pandas as pd

SPELLING_BEE_WORD_RE = re.compile(r"[a-z]{4,}")


class CorpusWordsWrapper:
    """
    Wrapper class for an NLTK corpus that deduplicates words and filters out
    unacceptable words (i.e. too short, or non-ascii characters).
    """

    def __init__(self, corpus: nltk.corpus.CorpusReader):
        """
        Initializer for CorpusWordsWrapper

        Args:
            corpus: the underlying corpus from NLTK
        """
        self.corpus = corpus
        self._words = None

    def words(self) -> tp.Tuple:
        """
        Get words from the corpus. Remove duplicates, words shorter than four
        letters, and words with non-ascii characters.

        Returns:
            Tuple: lowercase, deduplicated, filtered, sorted words
        """
        if self._words is None:
            words_lower = (word.lower() for word in self.corpus.words())
            words_dedup = set(words_lower)
            words_filtered = (word for word in words_dedup if SPELLING_BEE_WORD_RE.fullmatch(word) is not None)
            self._words = tuple(sorted(words_filtered))

        return self._words


CORPUS_NAMES = [
    "brown",
    "gutenberg",
    "inaugural",
    "reuters",
    "webtext",
    "words",
]
CORPORA = {k: CorpusWordsWrapper(getattr(nltk.corpus, k)) for k in CORPUS_NAMES}


@cache
def load_default_dictionary() -> tp.Tuple:
    """
    Load word lists from several NLTK corpora and combine to create a large word
    list.

    Returns:
        Tuple: tuple of all words
    """
    return tuple(sorted(set().union(*(corpus.words() for corpus in CORPORA.values()))))


class DictionaryDataFrameCache:
    """
    Class for caching the results of load_dictionary_dataframe
    """

    dictionary_dataframe: pd.DataFrame = None

    @classmethod
    def load_dictionary_dataframe(cls) -> pd.DataFrame:
        """
        Load word lists from several NLTK corpora and arrange in a dataframe that
        tracks which words are in which corpora.

        Returns:
            pandas.DataFrame: words as indicies, boolean column for each corpus
        """
        if cls.dictionary_dataframe is None:
            cls.dictionary_dataframe = cls._load_dictionary_dataframe()

        # defensive copy because DataFrames are mutable
        return cls.dictionary_dataframe.copy()

    @classmethod
    def _load_dictionary_dataframe(cls) -> pd.DataFrame:
        df = pd.DataFrame(index=load_default_dictionary())

        for corpus_name, corpus in CORPORA.items():
            df[corpus_name] = False
            df[corpus_name][list(corpus.words())] = True

        return df


load_dictionary_dataframe = DictionaryDataFrameCache.load_dictionary_dataframe



import spelling_bee_cheat.archive as archive
import spelling_bee_cheat.data_model as data_model
import spelling_bee_cheat.dictionary as dictionary
import spelling_bee_cheat.folders as folders
import spelling_bee_cheat.markov_search as markov_search
import spelling_bee_cheat.scrape as scrape
from spelling_bee_cheat.dictionary_search import dictionary_search
from spelling_bee_cheat.nyt_scraping import HintData, PuzzleData



# standard libraries
import typing as tp

# third party libraries
from pydantic import BaseModel


class Puzzle(BaseModel):
    """
    Representation of the JSON puzzle data used by NYT.
    """

    expiration: int = None
    displayWeekday: str
    displayDate: str
    printDate: str
    centerLetter: str
    outerLetters: tp.List[str]
    validLetters: tp.List[str]
    pangrams: tp.List[str]
    answers: tp.List[str]
    id: int
    freeExpiration: int
    editor: str



#!/usr/bin/python
# third party libraries
import click

import spelling_bee_cheat


@click.command()
def markov():
    """
    Download today's spelling bee puzzle from the NYT website and suggest words
    based on a character level Markov model.
    """
    pdat = spelling_bee_cheat.PuzzleData.fetch_today()
    ms = spelling_bee_cheat.markov_search.MarkovSearch()
    words = ms.markov_search(pdat)
    click.echo_via_pager("".join(word + "\n" for word in words))


if __name__ == "__main__":
    markov()
